@@ -123,6 +123,14 @@ bool debSrcRecordParser::BuildDepends(std::vector<pkgSrcRecords::Parser::BuildDe
 
       while (1)
       {
+	 // Strip off leading spaces (is done by ParseDepends, too) and
+	 // superfluous commas (encountered in user-written dsc/control files)
+	 do {
+	    for (;Start != Stop && isspace_ascii(*Start) != 0; ++Start);
+	 } while (*Start == ',' && ++Start != Stop);
+	 if (Start == Stop)
+	    break;
+
 	 BuildDepRec rec;
 	 Start = debListParser::ParseDepends(Start, Stop,
 					     rec.Package, rec.Version, rec.Op, true, StripMultiArch, true);
@@ -135,19 +143,12 @@ bool debSrcRecordParser::BuildDepends(std::vector<pkgSrcRecords::Parser::BuildDe
 	 // or something).
 	 if (rec.Package.empty())
 	 {
-	    // If we are in an OR group, we need to set the "Or" flag of the
-	    // previous entry to our value.
-	    if (BuildDeps.empty() == false && (BuildDeps[BuildDeps.size() - 1].Op & pkgCache::Dep::Or) == pkgCache::Dep::Or)
-	    {
-	       BuildDeps[BuildDeps.size() - 1].Op &= ~pkgCache::Dep::Or;
-	       BuildDeps[BuildDeps.size() - 1].Op |= (rec.Op & pkgCache::Dep::Or);
-	    }
+	    // If this was the last or-group member, close the or-group with the previous entry
+	    if (not BuildDeps.empty() && (BuildDeps.back().Op & pkgCache::Dep::Or) == pkgCache::Dep::Or && (rec.Op & pkgCache::Dep::Or) != pkgCache::Dep::Or)
+	       BuildDeps.back().Op &= ~pkgCache::Dep::Or;
 	 } else {
 	    BuildDeps.emplace_back(std::move(rec));
 	 }
-
-	 if (Start == Stop)
-	    break;
       }
    }
 
@@ -158,30 +159,7 @@ bool debSrcRecordParser::BuildDepends(std::vector<pkgSrcRecords::Parser::BuildDe
 // ---------------------------------------------------------------------
 /* This parses the list of files and returns it, each file is required to have
    a complete source package */
-bool debSrcRecordParser::Files(std::vector<pkgSrcRecords::File> &F)
-{
-   std::vector<pkgSrcRecords::File2> F2;
-   if (Files2(F2) == false)
-      return false;
-   for (std::vector<pkgSrcRecords::File2>::const_iterator f2 = F2.begin(); f2 != F2.end(); ++f2)
-   {
-      pkgSrcRecords::File2 f;
-#if __GNUC__ >= 4
-	#pragma GCC diagnostic push
-	#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
-#endif
-      f.MD5Hash = f2->MD5Hash;
-      f.Size = f2->Size;
-#if __GNUC__ >= 4
-	#pragma GCC diagnostic pop
-#endif
-      f.Path = f2->Path;
-      f.Type = f2->Type;
-      F.push_back(f);
-   }
-   return true;
-}
-bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
+bool debSrcRecordParser::Files(std::vector<pkgSrcRecords::File> &List)
 {
    List.clear();
 
@@ -231,7 +209,7 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	    path = Base + path;
 
 	 // look if we have a record for this file already
-	 std::vector<pkgSrcRecords::File2>::iterator file = List.begin();
+	 std::vector<pkgSrcRecords::File>::iterator file = List.begin();
 	 for (; file != List.end(); ++file)
 	    if (file->Path == path)
 	       break;
@@ -239,12 +217,6 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	 // we have it already, store the new hash and be done
 	 if (file != List.end())
 	 {
-	    if (checksumField == "Files")
-	    {
-	       APT_IGNORE_DEPRECATED_PUSH
-	       file->MD5Hash = hash;
-	       APT_IGNORE_DEPRECATED_POP
-	    }
 	    // an error here indicates that we have two different hashes for the same file
 	    if (file->Hashes.push_back(hashString) == false)
 	       return _error->Error("Error parsing checksum in %s of source package %s", checksumField.c_str(), Package().c_str());
@@ -252,18 +224,12 @@ bool debSrcRecordParser::Files2(std::vector<pkgSrcRecords::File2> &List)
 	 }
 
 	 // we haven't seen this file yet
-	 pkgSrcRecords::File2 F;
+	 pkgSrcRecords::File F;
 	 F.Path = path;
 	 F.FileSize = strtoull(size.c_str(), NULL, 10);
 	 F.Hashes.push_back(hashString);
 	 F.Hashes.FileSize(F.FileSize);
 
-	 APT_IGNORE_DEPRECATED_PUSH
-	 F.Size = F.FileSize;
-	 if (checksumField == "Files")
-	    F.MD5Hash = hash;
-	 APT_IGNORE_DEPRECATED_POP
-
 	 // Try to guess what sort of file it is we are getting.
 	 string::size_type Pos = F.Path.length()-1;
 	 while (1)
